@@ -15,19 +15,30 @@
 #include "nanopolish_anchor.h"
 #include "nanopolish_variant.h"
 
+#define MAX_EVENT_TO_BP_RATIO 20
+
 // structs
 struct SequenceAlignmentRecord
 {
+    SequenceAlignmentRecord(const bam1_t* record);
+
+    std::string read_name;
     std::string sequence;
     std::vector<AlignedPair> aligned_bases;
+    uint8_t rc; // with respect to reference genome
 };
 
 struct EventAlignmentRecord
 {
+    EventAlignmentRecord() {}
+    EventAlignmentRecord(SquiggleRead* sr,
+                         const int strand_idx,
+                         const SequenceAlignmentRecord& seq_record);
+
     SquiggleRead* sr;
     uint8_t rc; // with respect to reference genome
     uint8_t strand; // 0 = template, 1 = complement
-    uint8_t stride; // whether event indices increase or decrease along the reference
+    int8_t stride; // whether event indices increase or decrease along the reference
     std::vector<AlignedPair> aligned_events;
 };
 
@@ -40,17 +51,25 @@ class AlignmentDB
         AlignmentDB(const std::string& reads_file,
                     const std::string& reference_file,
                     const std::string& sequence_bam,
-                    const std::string& event_bam,
-                    const bool calibrate_reads = false);
+                    const std::string& event_bam);
 
         ~AlignmentDB();
 
         void load_region(const std::string& contig,
                          int start_position,
                          int stop_position);
+    
+        // Some high quality basecallers, like scrappie, may not output event
+        // annotations. This call is to support using scrappie basecalls
+        // with metrichor events.
+        void set_alternative_basecalls_bam(const std::string& alternative_sequence_bam);
         
         const std::string& get_reference() const { return m_region_ref_sequence; }
 
+        bool are_coordinates_valid(const std::string& contig,
+                                   int start_position,
+                                   int stop_position) const;
+
         std::string get_reference_substring(const std::string& contig,
                                             int start_position,
                                             int stop_position) const;
@@ -79,30 +98,35 @@ class AlignmentDB
         int get_region_end() const { return m_region_end; }
         
         void set_alternative_model_type(const std::string model_type_string) { m_model_type_string = model_type_string; }
-
-    private:
         
-        void _load_sequence_by_region();
-        void _load_events_by_region();
-        void _clear_region();
-
-        std::vector<EventAlignment> _build_event_alignment(const EventAlignmentRecord& event_record) const;
-
         // Search the vector of AlignedPairs using lower_bound/upper_bound
         // and the input reference coordinates. If the search succeeds,
         // set read_start/read_stop to be the read_pos of the bounding elements
         // and return true. 
-        bool _find_by_ref_bounds(const std::vector<AlignedPair>& pairs,
+        static bool _find_by_ref_bounds(const std::vector<AlignedPair>& pairs,
                                  int ref_start,
                                  int ref_stop,
                                  int& read_start,
-                                 int& read_stop) const;
+                                 int& read_stop);
 
-        bool _find_iter_by_ref_bounds(const std::vector<AlignedPair>& pairs,
+        static bool _find_iter_by_ref_bounds(const std::vector<AlignedPair>& pairs,
                                       int ref_start,
                                       int ref_stop,
                                       AlignedPairConstIter& start_iter,
-                                      AlignedPairConstIter& stop_iter) const;
+                                      AlignedPairConstIter& stop_iter);
+    private:
+        
+        std::vector<SequenceAlignmentRecord> _load_sequence_by_region(const std::string& sequence_bam);
+        std::vector<EventAlignmentRecord> _load_events_by_region_from_bam(const std::string& event_bam);
+        std::vector<EventAlignmentRecord> _load_events_by_region_from_read(const std::vector<SequenceAlignmentRecord>& sequence_records);
+        void _load_squiggle_read(const std::string& read_name);
+
+        void _clear_region();
+
+        void _debug_print_alignments();
+
+        std::vector<EventAlignment> _build_event_alignment(const EventAlignmentRecord& event_record) const;
+
 
         //
         // data
@@ -110,9 +134,9 @@ class AlignmentDB
         std::string m_reference_file;
         std::string m_sequence_bam;
         std::string m_event_bam;
+        std::string m_alternative_basecalls_bam;
 
         // parameters
-        bool m_calibrate_on_load;
 
         // loaded region
         std::string m_region_ref_sequence;
@@ -121,7 +145,7 @@ class AlignmentDB
         int m_region_end;
 
         // cached alignments for a region
-        Fast5Map m_fast5_name_map;
+        ReadDB m_read_db;
         std::vector<SequenceAlignmentRecord> m_sequence_records;
         std::vector<EventAlignmentRecord> m_event_records;
         SquiggleReadMap m_squiggle_read_map;
