@@ -12,6 +12,7 @@
 #include <cstring>
 #include <inttypes.h>
 #include <assert.h>
+#include <algorithm>
 #include "nanopolish_iupac.h"
 
 #define METHYLATED_SYMBOL 'M'
@@ -63,7 +64,7 @@ class Alphabet
         virtual char base(uint8_t r) const = 0;
         virtual char complement(char b) const = 0;
         virtual uint32_t size() const = 0;
-        virtual std::string get_name() const = 0;
+        virtual const char* get_name() const = 0;
 
         // support for methylated bases with recognition sequences
         virtual size_t num_recognition_sites() const = 0;
@@ -151,7 +152,10 @@ class Alphabet
         // return a new copy of the string with IUPAC ambiguity characters changed
         virtual std::string disambiguate(const std::string& str) const
         {
+            // create output and convert lower case to upper case
             std::string out(str);
+            std::transform(out.begin(), out.end(), out.begin(), ::toupper);
+
             size_t i = 0;
             while(i < out.length()) {
                 size_t stride = 1;
@@ -160,7 +164,7 @@ class Alphabet
                 // Does this location (partially) match a methylated recognition site?
                 for(size_t j = 0; j < num_recognition_sites(); ++j) {
 
-                    RecognitionMatch match = match_to_site(str, i, get_recognition_site_methylated(j), recognition_length());
+                    RecognitionMatch match = match_to_site(out, i, get_recognition_site_methylated(j), recognition_length());
                     if(match.length > 0) {
                         stride = match.length; // skip to end of match
                         is_recognition_site = true;
@@ -234,7 +238,7 @@ class Alphabet
         }
 
         // does this alphabet contain all of the nucleotides in bases?
-        virtual inline bool contains_all(const char *bases) const = 0;
+        virtual bool contains_all(const char *bases) const = 0;
 };
 
 #define BASIC_MEMBER_BOILERPLATE \
@@ -245,7 +249,7 @@ class Alphabet
     static const uint32_t _size; 
 
 #define BASIC_ACCESSOR_BOILERPLATE \
-    virtual std::string get_name() const { return _name; } \
+    virtual const char* get_name() const { return _name; } \
     virtual uint8_t rank(char b) const { return _rank[(int)b]; }        \
     virtual char base(uint8_t r) const { return _base[r]; } \
     virtual char complement(char b) const { return _complement[_rank[(int)b]]; } \
@@ -275,6 +279,34 @@ struct DNAAlphabet : public Alphabet
     }
 };
 
+// An RNA alphabet where the Us have been converted to Ts
+// This is used for direct RNA sequencing as most reference sequences
+// contain Ts into of U
+struct UtoTRNAAlphabet : public Alphabet
+{
+    // members
+    BASIC_MEMBER_BOILERPLATE
+
+    // functions
+    BASIC_ACCESSOR_BOILERPLATE
+
+    // no methylation in this alphabet
+    virtual size_t num_recognition_sites() const { return 0; }
+    virtual size_t recognition_length() const { return 0; }
+    virtual const char* get_recognition_site(size_t) const { return NULL; }
+    virtual const char* get_recognition_site_methylated(size_t) const { return NULL; }
+    virtual const char* get_recognition_site_methylated_complement(size_t) const { 
+        return NULL;
+    }
+
+    // does this alphabet contain all of the nucleotides in bases?
+    virtual inline bool contains_all(const char *bases) const
+    {
+        return strspn(bases, _base) == strlen(bases);
+    }
+};
+
+
 #define METHYLATION_MEMBER_BOILERPLATE \
     static const uint32_t _num_recognition_sites; \
     static const uint32_t _recognition_length; \
@@ -356,7 +388,9 @@ extern DNAAlphabet gDNAAlphabet;
 extern MethylCpGAlphabet gMCpGAlphabet;
 extern MethylDamAlphabet gMethylDamAlphabet;
 extern MethylDcmAlphabet gMethylDcmAlphabet;
+extern UtoTRNAAlphabet gUtoTRNAAlphabet;
 
+std::vector<const Alphabet*> get_alphabet_list();
 const Alphabet* best_alphabet(const char *bases);
 const Alphabet* get_alphabet_by_name(const std::string& name);
 
