@@ -32,9 +32,9 @@
 #include "nanopolish_matrix.h"
 #include "nanopolish_profile_hmm.h"
 #include "nanopolish_anchor.h"
-#include "nanopolish_fast5_map.h"
 #include "nanopolish_model_names.h"
 #include "nanopolish_pore_model_set.h"
+#include "nanopolish_read_db.h"
 #include "training_core.hpp"
 #include "H5pubconf.h"
 #include "profiler.h"
@@ -42,7 +42,7 @@
 #include "logger.hpp"
 
 #include "nanopolish_scorereads.h"
-#include "../eigen/Eigen/Dense"
+#include "Eigen/Dense"
 
 extern float g_p_skip, g_p_skip_self, g_p_bad, g_p_bad_self;
 
@@ -105,10 +105,11 @@ static const char *METHYLTRAIN_USAGE_MESSAGE =
 "  -b, --bam=FILE                       the reads aligned to the genome assembly are in bam FILE\n"
 "  -g, --genome=FILE                    the reference genome is in FILE\n"
 "  -t, --threads=NUM                    use NUM threads (default: 1)\n"
-"      --filter-policy=STR              filter reads for [R7-methylation] or [R9-nucleotide] project\n"
+"      --filter-policy=STR              filter reads for [R7] or [R9] project\n"
 "  -s, --out-suffix=STR                 name output files like <strand>.out_suffix\n"
 "      --out-fofn=FILE                  write the names of the output models into FILE\n"
 "      --rounds=NUM                     number of training rounds to perform\n"
+"      --max-reads=NUM                  stop after processing NUM reads in each round\n"
 "      --progress                       print out a progress message\n"
 "      --stdv                           enable stdv modelling\n"
 "\nReport bugs to " PACKAGE_BUGREPORT "\n\n";
@@ -133,6 +134,7 @@ namespace opt
     static unsigned progress = 0;
     static unsigned num_threads = 1;
     static unsigned batch_size = 128;
+    static unsigned max_reads = -1;
 
     // Constants that determine which events to use for training
     static float min_event_duration = 0.002;
@@ -157,7 +159,8 @@ enum { OPT_HELP = 1,
        OPT_P_SKIP,
        OPT_P_SKIP_SELF,
        OPT_P_BAD,
-       OPT_P_BAD_SELF
+       OPT_P_BAD_SELF,
+       OPT_MAX_READS
      };
 
 static const struct option longopts[] = {
@@ -184,6 +187,7 @@ static const struct option longopts[] = {
     { "log-level",          required_argument, NULL, OPT_LOG_LEVEL },
     { "filter-policy",      required_argument, NULL, OPT_FILTER_POLICY },
     { "rounds",             required_argument, NULL, OPT_NUM_ROUNDS },
+    { "max-reads",          required_argument, NULL, OPT_MAX_READS },
     { NULL, 0, NULL, 0 }
 };
 
@@ -191,14 +195,14 @@ static const struct option longopts[] = {
 // returns true if the recalibration was performed
 // in either case, sets residual to the L1 norm of the residual
 bool recalibrate_model(SquiggleRead &sr,
+                       const PoreModel& pore_model,
                        const int strand_idx,
                        const std::vector<EventAlignment> &alignment_output,
-                       const Alphabet* alphabet,
-                       const bool scale_var, 
+                       const bool scale_var,
                        const bool scale_drift)
 {
     std::vector<double> raw_events, times, level_means, level_stdvs;
-    uint32_t k = sr.pore_model[strand_idx].k;
+    uint32_t k = pore_model.k;
     const uint32_t num_equations = scale_drift ? 3 : 2;
 
     //std::cout << "Previous pore model parameters: " << sr.pore_model[strand_idx].shift << ", "
@@ -207,16 +211,23 @@ bool recalibrate_model(SquiggleRead &sr,
     //                                                << sr.pore_model[strand_idx].var   << std::endl;
 
     // extract necessary vectors from the read and the pore model; note do not want scaled values
-    for ( const auto &ea : alignment_output ) {
+    for(size_t ei = 0; ei < alignment_output.size(); ++ei) {
+        const auto& ea = alignment_output[ei];
         if(ea.hmm_state == 'M') {
-            std::string model_kmer = ea.rc ? alphabet->reverse_complement(ea.ref_kmer) : ea.ref_kmer;
-            uint32_t rank = alphabet->kmer_rank(model_kmer.c_str(), k);
+            std::string model_kmer = ea.rc ? pore_model.pmalphabet->reverse_complement(ea.ref_kmer) : ea.ref_kmer;
+            uint32_t rank = pore_model.pmalphabet->kmer_rank(model_kmer.c_str(), k);
 
-            raw_events.push_back ( sr.get_uncorrected_level(ea.event_idx, strand_idx) );
-            level_means.push_back( sr.pore_model[strand_idx].states[rank].level_mean );
-            level_stdvs.push_back( sr.pore_model[strand_idx].states[rank].level_stdv );
+            raw_events.push_back ( sr.get_unscaled_level(ea.event_idx, strand_idx) );
+            level_means.push_back( pore_model.states[rank].level_mean );
+            level_stdvs.push_back( pore_model.states[rank].level_stdv );
             if (scale_drift)
                 times.push_back  ( sr.get_time(ea.event_idx, strand_idx) );
+
+            /*
+            fprintf(stdout, "recalibrate ei: %zu level: %.2lf kmer: %s model: %.2lf\n", 
+                    ei, sr.get_uncorrected_level(ea.event_idx, strand_idx), model_kmer.c_str(), 
+                    sr.pore_model[strand_idx].states[rank].level_mean);
+            */
         }
     }
 
@@ -241,8 +252,8 @@ bool recalibrate_model(SquiggleRead &sr,
             double mu = level_means[i];
             double e  = raw_events[i];
 
-            A(0,0) += inv_var;  A(0,1) += mu*inv_var;    
-                                A(1,1) += mu*mu*inv_var; 
+            A(0,0) += inv_var;  A(0,1) += mu*inv_var;
+                                A(1,1) += mu*mu*inv_var;
 
             b(0) += e*inv_var;
             b(1) += mu*e*inv_var;
@@ -267,13 +278,10 @@ bool recalibrate_model(SquiggleRead &sr,
         double shift = x(0);
         double scale = x(1);
         double drift = scale_drift ? x(2) : 0.;
-
-        sr.pore_model[strand_idx].shift = shift;
-        sr.pore_model[strand_idx].scale = scale;
-        sr.pore_model[strand_idx].drift = drift;
+        double var = 1.0;
 
         if (scale_var) {
-            double var = 0.;
+            var = 0.;
             for (size_t i=0; i<raw_events.size(); i++) {
                 double yi = (raw_events[i] - shift - scale*level_means[i]);
                 if (scale_drift)
@@ -281,53 +289,47 @@ bool recalibrate_model(SquiggleRead &sr,
                 var+= yi*yi/(level_stdvs[i]*level_stdvs[i]);
             }
             var /= raw_events.size();
-
-            sr.pore_model[strand_idx].var   = sqrt(var); // 'var' is really the scaling for std dev.
+            var = sqrt(var);
         }
 
-        if (sr.pore_model[strand_idx].is_scaled)
-            sr.pore_model[strand_idx].bake_gaussian_parameters();
-
+        sr.scalings[strand_idx].set4(shift, scale, drift, var);
         recalibrated = true;
-        //std::cout << "Updated pore model parameters:  " << sr.pore_model[strand_idx].shift << ", "
-        //                                                << sr.pore_model[strand_idx].scale << ", "
-        //                                                << sr.pore_model[strand_idx].drift << ", "
-        //                                                << sr.pore_model[strand_idx].var   << std::endl;
     }
-        
+
     return recalibrated;
 }
 
 // Update the training data with aligned events from a read
-void add_aligned_events(const Fast5Map& name_map,
+void add_aligned_events(const ReadDB& read_db,
                         const faidx_t* fai,
                         const bam_hdr_t* hdr,
                         const bam1_t* record,
                         size_t read_idx,
                         int region_start,
                         int region_end,
+                        const std::string& training_kit,
+                        const std::string& training_alphabet,
+                        size_t training_k,
                         size_t round,
                         ModelTrainingMap& training)
 {
     // Load a squiggle read for the mapped read
     std::string read_name = bam_get_qname(record);
-    std::string fast5_path = name_map.get_path(read_name);
 
     // load read
-    SquiggleRead sr(read_name, fast5_path);
-
-    // replace the models that are built into the read with the current trained model
-    sr.replace_models(opt::trained_model_type);
+    SquiggleRead sr(read_name, read_db);
 
     for(size_t strand_idx = 0; strand_idx < NUM_STRANDS; ++strand_idx) {
-
         // skip if 1D reads and this is the wrong strand
         if(!sr.has_events_for_strand(strand_idx)) {
             continue;
         }
 
+        assert(training_kit == sr.get_model_kit_name(strand_idx));
+        assert(training_k == sr.get_model_k(strand_idx));
+
         // set k
-        uint32_t k = sr.pore_model[strand_idx].k;
+        uint32_t k = sr.get_model_k(strand_idx);
 
         // Align to the new model
         EventAlignmentParameters params;
@@ -337,41 +339,49 @@ void add_aligned_events(const Fast5Map& name_map,
         params.record = record;
         params.strand_idx = strand_idx;
 
-        params.alphabet = mtrain_alphabet;
+        params.alphabet = mtrain_alphabet->get_name();
         params.read_idx = read_idx;
         params.region_start = region_start;
         params.region_end = region_end;
+
         std::vector<EventAlignment> alignment_output = align_read_to_ref(params);
         if (alignment_output.size() == 0)
             return;
 
         // Update pore model based on alignment
-        std::string curr_model = sr.pore_model[strand_idx].metadata.get_short_name();
-        double orig_score = -INFINITY;
+        std::string model_key = PoreModelSet::get_model_key(*sr.get_model(strand_idx, mtrain_alphabet->get_name()));
 
+        //
+        // Optional recalibration of shift/scale/drift and output of sequence likelihood
+        //
+        double orig_score = -INFINITY;
         if (opt::output_scores) {
             orig_score = model_score(sr, strand_idx, fai, alignment_output, 500, NULL);
 
             #pragma omp critical(print)
-            std::cout << round << " " << curr_model << " " << read_idx << " " << strand_idx << " Original " << orig_score << std::endl;
+            std::cout << round << " " << model_key << " " << read_idx << " " << strand_idx << " Original " << orig_score << std::endl;
         }
 
         if ( opt::calibrate ) {
             double resid = 0.;
-            recalibrate_model(sr, strand_idx, alignment_output, mtrain_alphabet, resid, true);
+            recalibrate_model(sr, *sr.get_model(strand_idx, mtrain_alphabet->get_name()), strand_idx, alignment_output, resid, true);
 
             if (opt::output_scores) {
                 double rescaled_score = model_score(sr, strand_idx, fai, alignment_output, 500, NULL);
                 #pragma omp critical(print)
                 {
-                    std::cout << round << " " << curr_model << " " << read_idx << " " << strand_idx << " Rescaled " << rescaled_score << std::endl;
-                    std::cout << round << " " << curr_model << " " << read_idx << " " << strand_idx << " Delta " << rescaled_score-orig_score << std::endl;
+                    std::cout << round << " " << model_key << " " << read_idx << " " << strand_idx << " Rescaled " << rescaled_score << std::endl;
+                    std::cout << round << " " << model_key << " " << read_idx << " " << strand_idx << " Delta " << rescaled_score-orig_score << std::endl;
                 }
             }
         }
+        //
+        // Get the data structure holding the training data for this strand
+        //
+        auto emission_map_iter = training.find(model_key);
 
-        // Get the training data for this model
-        auto& emission_map = training[curr_model];
+        assert(emission_map_iter != training.end());
+        auto& emission_map = emission_map_iter->second;
 
         for(size_t i = 0; i < alignment_output.size(); ++i) {
             const EventAlignment& ea = alignment_output[i];
@@ -450,7 +460,7 @@ void parse_methyltrain_options(int argc, char** argv)
             case 's': arg >> opt::out_suffix; break;
             case 'v': opt::verbose++; break;
             case 'c': opt::calibrate = 1; break;
-            case OPT_STDV: model_stdv() = true; break;
+            case OPT_STDV: /*model_stdv() = true;*/ break;
             case OPT_OUT_FOFN: arg >> opt::out_fofn; break;
             case OPT_NUM_ROUNDS: arg >> opt::num_training_rounds; break;
             case OPT_OUTPUT_SCORES: opt::output_scores = true; break;
@@ -462,6 +472,7 @@ void parse_methyltrain_options(int argc, char** argv)
             case OPT_P_SKIP_SELF: arg >> g_p_skip_self; break;
             case OPT_P_BAD: arg >> g_p_bad; break;
             case OPT_P_BAD_SELF: arg >> g_p_bad_self; break;
+            case OPT_MAX_READS: arg >> opt::max_reads; break;
             case OPT_HELP:
                 std::cout << METHYLTRAIN_USAGE_MESSAGE;
                 exit(EXIT_SUCCESS);
@@ -469,7 +480,7 @@ void parse_methyltrain_options(int argc, char** argv)
                 std::cout << METHYLTRAIN_VERSION_MESSAGE;
                 exit(EXIT_SUCCESS);
             case OPT_LOG_LEVEL:
-                Logger::set_level_from_option(arg.str());
+                logger::Logger::set_level_from_option(arg.str());
                 break;
         }
     }
@@ -492,7 +503,7 @@ void parse_methyltrain_options(int argc, char** argv)
         std::cerr << SUBPROGRAM ": a --reads file must be provided\n";
         die = true;
     }
-    
+
     if(opt::genome_file.empty()) {
         std::cerr << SUBPROGRAM ": a --genome file must be provided\n";
         die = true;
@@ -503,14 +514,6 @@ void parse_methyltrain_options(int argc, char** argv)
         die = true;
     }
 
-    if(opt::models_fofn.empty()) {
-        std::cerr << SUBPROGRAM ": a --models file must be provided\n";
-        die = true;
-    } else {
-        // initialize the model set from the fofn
-        PoreModelSet::initialize(opt::models_fofn);
-    }
-
     // Parse the training target string
     if(training_target_str != "") {
         if(training_target_str == "unmethylated") {
@@ -527,9 +530,12 @@ void parse_methyltrain_options(int argc, char** argv)
 
     // Parse the training target string
     if(filter_policy_str != "") {
-        if(filter_policy_str == "R9-nucleotide") {
-        } else if(filter_policy_str == "R7-methylation") {
-            // default, do nothing
+        if(filter_policy_str == "R9") {
+            opt::min_event_duration = 0.002;
+        } else if(filter_policy_str == "R9.4") {
+            opt::min_event_duration = 0.001;
+        } else if(filter_policy_str == "R7") {
+            opt::min_event_duration = 0.005;
         } else {
             std::cerr << SUBPROGRAM ": unknown --filter-policy\n";
             die = true;
@@ -542,9 +548,154 @@ void parse_methyltrain_options(int argc, char** argv)
     }
 }
 
-void train_one_round(const Fast5Map& name_map, size_t round)
+struct TrainingResult
 {
-    const PoreModelMap& current_models = PoreModelSet::get_models(opt::trained_model_type);
+    PoreModel trained_model;
+    size_t num_kmers_trained;
+};
+
+
+TrainingResult retrain_model_from_events(const PoreModel& current_model,
+                                         const std::vector<StateSummary>& summaries,
+                                         std::ofstream& training_ofs,
+                                         FILE* summary_fp)
+{
+    TrainingResult result;
+
+    // Initialize the new model from the current model
+    result.trained_model = current_model;
+    result.num_kmers_trained = 0;
+
+    uint32_t k = result.trained_model.k;
+    std::string model_key = PoreModelSet::get_model_key(result.trained_model);
+    std::string model_short_name = result.trained_model.metadata.get_short_name();
+
+    // Generate the complete set of kmers
+    std::string gen_kmer(k, 'A');
+    std::vector<std::string> all_kmers;
+    for(size_t ki = 0; ki < summaries.size(); ++ki) {
+        all_kmers.push_back(gen_kmer);
+        mtrain_alphabet->lexicographic_next(gen_kmer);
+    }
+    assert(gen_kmer == std::string(k, 'A'));
+    assert(all_kmers.front() == std::string(k, 'A'));
+    assert(all_kmers.back() == std::string(k, 'T'));
+
+    // Update means for each kmer
+    #pragma omp parallel for
+    for(size_t ki = 0; ki < summaries.size(); ++ki) {
+        assert(ki < all_kmers.size());
+        std::string kmer = all_kmers[ki];
+
+        // write the observed values to a tsv file
+        #pragma omp critical
+        {
+            for(size_t ei = 0; ei < summaries[ki].events.size(); ++ei) {
+                summaries[ki].events[ei].write_tsv(training_ofs, model_short_name, kmer);
+            }
+        }
+
+        bool is_m_kmer = kmer.find('M') != std::string::npos;
+        bool update_kmer = opt::training_target == TT_ALL_KMERS ||
+                           (is_m_kmer && opt::training_target == TT_METHYLATED_KMERS) ||
+                           (!is_m_kmer && opt::training_target == TT_UNMETHYLATED_KMERS);
+
+        bool trained = false;
+
+        // only train if there are a sufficient number of events for this kmer
+        if(update_kmer && summaries[ki].events.size() >= opt::min_number_of_events_to_train) {
+
+            // train a mixture model where a minority of k-mers aren't methylated
+            ParamMixture mixture;
+
+            float incomplete_methylation_rate = 0.05f;
+            std::string um_kmer = mtrain_alphabet->unmethylate(kmer);
+            size_t um_ki = mtrain_alphabet->kmer_rank(um_kmer.c_str(), k);
+
+            // Initialize the training parameters. If this is a kmer containing
+            // a methylation site we train a two component mixture, otherwise
+            // just fit a gaussian
+            float major_weight = is_m_kmer ? 1 - incomplete_methylation_rate : 1.0f;
+            mixture.log_weights.push_back(log(major_weight));
+            mixture.params.push_back(current_model.get_parameters(ki));
+
+            if(is_m_kmer) {
+                // add second unmethylated component
+                mixture.log_weights.push_back(std::log(incomplete_methylation_rate));
+                mixture.params.push_back(current_model.get_parameters(um_ki));
+            }
+
+            if(opt::verbose > 1) {
+                fprintf(stderr, "INIT__MIX %s\t%s\t[%.2lf %.2lf %.2lf]\t[%.2lf %.2lf %.2lf]\n", model_key.c_str(), kmer.c_str(),
+                    std::exp(mixture.log_weights[0]), mixture.params[0].level_mean, mixture.params[0].level_stdv,
+                    std::exp(mixture.log_weights[1]), mixture.params[1].level_mean, mixture.params[1].level_stdv);
+            }
+
+            ParamMixture trained_mixture = train_gaussian_mixture(summaries[ki].events, mixture);
+
+            if(opt::verbose > 1) {
+                fprintf(stderr, "TRAIN_MIX %s\t%s\t[%.2lf %.2lf %.2lf]\t[%.2lf %.2lf %.2lf]\n", model_key.c_str(), kmer.c_str(),
+                    std::exp(trained_mixture.log_weights[0]), trained_mixture.params[0].level_mean, trained_mixture.params[0].level_stdv,
+                    std::exp(trained_mixture.log_weights[1]), trained_mixture.params[1].level_mean, trained_mixture.params[1].level_stdv);
+            }
+
+            #pragma omp critical
+            result.trained_model.states[ki] = trained_mixture.params[0];
+
+#if 0
+            if (false && model_stdv()) {
+                ParamMixture ig_mixture;
+                // weights
+                ig_mixture.log_weights = trained_mixture.log_weights;
+                // states
+                ig_mixture.params.emplace_back(trained_mixture.params[0]);
+
+                if(is_m_kmer) {
+                    ig_mixture.params.emplace_back(current_model.get_parameters(um_ki));
+                }
+                // run training
+                auto trained_ig_mixture = train_invgaussian_mixture(summaries[ki].events, ig_mixture);
+
+                LOG("methyltrain", debug)
+                    << "IG_INIT__MIX " << model_key << " " << kmer.c_str() << " ["
+                    << std::fixed << std::setprecision(5) << ig_mixture.params[0].sd_mean << " "
+                    << ig_mixture.params[1].sd_mean << "]" << std::endl
+                    << "IG_TRAIN_MIX " << model_key << " " << kmer.c_str() << " ["
+                    << trained_ig_mixture.params[0].sd_mean << " "
+                    << trained_ig_mixture.params[1].sd_mean << "]" << std::endl;
+
+                // update state
+                #pragma omp critical
+                {
+                    result.trained_model.states[ki] = trained_ig_mixture.params[0];
+                }
+            }
+#endif 
+            #pragma omp atomic
+            result.num_kmers_trained += 1;
+        }
+
+        #pragma omp critical
+        {
+            fprintf(summary_fp, "%s\t%s\t%d\t%d\t%d\t%zu\t%d\t%.2lf\t%.2lf\n",
+                                    model_short_name.c_str(), kmer.c_str(),
+                                    summaries[ki].num_matches, summaries[ki].num_skips, summaries[ki].num_stays,
+                                    summaries[ki].events.size(), trained, result.trained_model.states[ki].level_mean, result.trained_model.states[ki].level_stdv);
+        }
+    }
+
+    return result;
+}
+
+void train_one_round(const ReadDB& read_db,
+                     const std::string& kit_name,
+                     const std::string& alphabet,
+                     size_t k,
+                     size_t round)
+{
+
+    // Get a copy of the models for each strand for this datatype
+    const std::map<std::string, PoreModel> current_models = PoreModelSet::copy_strand_models(kit_name, alphabet, k);
 
     // Initialize the training summary stats for each kmer for each model
     ModelTrainingMap model_training_data;
@@ -607,19 +758,22 @@ void train_one_round(const Fast5Map& name_map, size_t round)
 
     do {
         assert(num_records_buffered < records.size());
-        
+
         // read a record into the next slot in the buffer
         result = sam_itr_next(bam_fh, itr, records[num_records_buffered]);
         num_records_buffered += result >= 0;
 
         // realign if we've hit the max buffer size or reached the end of file
-        if(num_records_buffered == records.size() || result < 0) {
-            #pragma omp parallel for            
+        if(num_records_buffered == records.size() || result < 0 || (num_records_buffered + num_reads_realigned == opt::max_reads)) {
+            #pragma omp parallel for
             for(size_t i = 0; i < num_records_buffered; ++i) {
                 bam1_t* record = records[i];
                 size_t read_idx = num_reads_realigned + i;
                 if( (record->core.flag & BAM_FUNMAP) == 0) {
-                    add_aligned_events(name_map, fai, hdr, record, read_idx, clip_start, clip_end, round, model_training_data);
+                    add_aligned_events(read_db, fai, hdr, record, read_idx,
+                                       clip_start, clip_end,
+                                       kit_name, alphabet, k,
+                                       round, model_training_data);
                 }
             }
 
@@ -630,11 +784,11 @@ void train_one_round(const Fast5Map& name_map, size_t round)
         if(opt::progress) {
             fprintf(stderr, "Realigned %zu reads in %.1lfs\r", num_reads_realigned, progress.get_elapsed_seconds());
         }
-    } while(result >= 0);
-    
+    } while(result >= 0 && num_reads_realigned < opt::max_reads);
+
     assert(num_records_buffered == 0);
     progress.end();
-    
+
     // open the summary file
     std::stringstream summary_fn;
     summary_fn << "methyltrain" << opt::out_suffix << ".summary";
@@ -652,133 +806,23 @@ void train_one_round(const Fast5Map& name_map, size_t round)
     StateTrainingData::write_header(training_ofs);
 
     // iterate over models: template, complement_pop1, complement_pop2
-    for(auto model_training_iter = model_training_data.begin(); 
-             model_training_iter != model_training_data.end(); model_training_iter++) {
-        
+    for(auto model_training_iter = model_training_data.begin();
+             model_training_iter != model_training_data.end(); model_training_iter++)
+    {
         // Initialize the trained model from the input model
         auto current_model_iter = current_models.find(model_training_iter->first);
         assert(current_model_iter != current_models.end());
-
-        std::string model_name = model_training_iter->first;
-        std::string model_short_name = current_model_iter->second.metadata.get_short_name();
-        
-        // Initialize the new model from the current model
-        PoreModel updated_model = current_model_iter->second;
-        uint32_t k = updated_model.k;
         const std::vector<StateSummary>& summaries = model_training_iter->second;
 
-        // Generate the complete set of kmers
-        std::string gen_kmer(k, 'A');
-        std::vector<std::string> all_kmers;
-        for(size_t ki = 0; ki < summaries.size(); ++ki) {
-            all_kmers.push_back(gen_kmer);
-            mtrain_alphabet->lexicographic_next(gen_kmer);
-        }
-        assert(gen_kmer == std::string(k, 'A'));
-        assert(all_kmers.front() == std::string(k, 'A'));
-        assert(all_kmers.back() == std::string(k, 'T'));
+        TrainingResult result = retrain_model_from_events(current_model_iter->second, summaries, training_ofs, summary_fp);
 
-        // Update means for each kmer
-        #pragma omp parallel for
-        for(size_t ki = 0; ki < summaries.size(); ++ki) {
-            assert(ki < all_kmers.size());
-            std::string kmer = all_kmers[ki];
+        // add the updated model into the collection (or replace what is already there)
+        PoreModelSet::add_model(result.trained_model);
 
-            // write the observed values to a tsv file
-            #pragma omp critical
-            {
-                for(size_t ei = 0; ei < summaries[ki].events.size(); ++ei) {
-                    summaries[ki].events[ei].write_tsv(training_ofs, model_short_name, kmer);
-                }
-
-            }
-
-            bool is_m_kmer = kmer.find('M') != std::string::npos;
-            bool update_kmer = opt::training_target == TT_ALL_KMERS ||
-                               (is_m_kmer && opt::training_target == TT_METHYLATED_KMERS) ||
-                               (!is_m_kmer && opt::training_target == TT_UNMETHYLATED_KMERS);
-            bool trained = false;
-            // only train if there are a sufficient number of events for this kmer
-            if(update_kmer && summaries[ki].events.size() >= opt::min_number_of_events_to_train) {
-                
-                // train a mixture model where a minority of k-mers aren't methylated
-                ParamMixture mixture;
-                
-                float incomplete_methylation_rate = 0.05f;
-                std::string um_kmer = mtrain_alphabet->unmethylate(kmer);
-                size_t um_ki = mtrain_alphabet->kmer_rank(um_kmer.c_str(), k);
-                
-                // Initialize the training parameters. If this is a kmer containing
-                // a methylation site we train a two component mixture, otherwise
-                // just fit a gaussian
-                float major_weight = is_m_kmer ? 1 - incomplete_methylation_rate : 1.0f;
-                mixture.log_weights.push_back(log(major_weight));
-                mixture.params.push_back(current_model_iter->second.get_parameters(ki));
-                
-                if(is_m_kmer) {
-                    // add second unmethylated component
-                    mixture.log_weights.push_back(std::log(incomplete_methylation_rate));
-                    mixture.params.push_back(current_model_iter->second.get_parameters(um_ki));
-                }
-
-                if(opt::verbose > 1) {
-                    fprintf(stderr, "INIT__MIX %s\t%s\t[%.2lf %.2lf %.2lf]\t[%.2lf %.2lf %.2lf]\n", model_training_iter->first.c_str(), kmer.c_str(), 
-                        std::exp(mixture.log_weights[0]), mixture.params[0].level_mean, mixture.params[0].level_stdv,
-                        std::exp(mixture.log_weights[1]), mixture.params[1].level_mean, mixture.params[1].level_stdv);
-                }
-
-                ParamMixture trained_mixture = train_gaussian_mixture(summaries[ki].events, mixture);
-
-                if(opt::verbose > 1) {
-                    fprintf(stderr, "TRAIN_MIX %s\t%s\t[%.2lf %.2lf %.2lf]\t[%.2lf %.2lf %.2lf]\n", model_training_iter->first.c_str(), kmer.c_str(), 
-                        std::exp(trained_mixture.log_weights[0]), trained_mixture.params[0].level_mean, trained_mixture.params[0].level_stdv,
-                        std::exp(trained_mixture.log_weights[1]), trained_mixture.params[1].level_mean, trained_mixture.params[1].level_stdv);
-                }
-
-                #pragma omp critical
-                updated_model.states[ki] = trained_mixture.params[0];
-
-                if (model_stdv()) {
-                    ParamMixture ig_mixture;
-                    // weights
-                    ig_mixture.log_weights = trained_mixture.log_weights;
-                    // states
-                    ig_mixture.params.emplace_back(trained_mixture.params[0]);
-
-                    if(is_m_kmer) {
-                        ig_mixture.params.emplace_back(current_model_iter->second.get_parameters(um_ki));
-                    }
-                    // run training
-                    auto trained_ig_mixture = train_invgaussian_mixture(summaries[ki].events, ig_mixture);
-
-                    LOG("methyltrain", debug)
-                        << "IG_INIT__MIX " << model_training_iter->first.c_str() << " " << kmer.c_str() << " ["
-                        << std::fixed << std::setprecision(5) << ig_mixture.params[0].sd_mean << " "
-                        << ig_mixture.params[1].sd_mean << "]" << std::endl
-                        << "IG_TRAIN_MIX " << model_training_iter->first.c_str() << " " << kmer.c_str() << " ["
-                        << trained_ig_mixture.params[0].sd_mean << " "
-                        << trained_ig_mixture.params[1].sd_mean << "]" << std::endl;
-
-                    // update state
-                    #pragma omp critical
-                    {
-                        updated_model.states[ki] = trained_ig_mixture.params[0];
-                    }
-                }
-
-                trained = true;
-            }
-
-            #pragma omp critical
-            {
-                fprintf(summary_fp, "%s\t%s\t%d\t%d\t%d\t%zu\t%d\t%.2lf\t%.2lf\n",
-                                        model_short_name.c_str(), kmer.c_str(), 
-                                        summaries[ki].num_matches, summaries[ki].num_skips, summaries[ki].num_stays, 
-                                        summaries[ki].events.size(), trained, updated_model.states[ki].level_mean, updated_model.states[ki].level_stdv);
-            }
-
-            // add the updated model into the collection (or replace what is already there)
-            PoreModelSet::insert_model(opt::trained_model_type, updated_model);
+        // write the updated model to disk
+        if(opt::write_models && result.num_kmers_trained > 0) {
+            std::string out_name = PoreModelSet::get_model_key(result.trained_model) + ".model";
+            result.trained_model.write(out_name, out_name);
         }
     }
 
@@ -796,25 +840,20 @@ void train_one_round(const Fast5Map& name_map, size_t round)
     fclose(summary_fp);
 }
 
-void write_models(const PoreModelMap& models, int round)
+void write_models(const std::map<std::string, PoreModel>& models, int round)
 {
-    // file-of-filenames containing the new models
-    std::ofstream fofn_writer(opt::out_fofn);
-
     // Write the model
-    for(auto model_iter = models.begin(); 
+    for(auto model_iter = models.begin();
              model_iter != models.end(); model_iter++) {
 
         assert(!model_iter->second.model_filename.empty());
         std::stringstream round_ss;
         round_ss << round;
 
-        std::string outname   =  model_iter->second.name + opt::out_suffix; 
+        std::string outname   =  model_iter->second.name + opt::out_suffix;
         std::string modelname =  model_iter->first + opt::out_suffix;
         model_iter->second.write( outname, modelname );
 
-        // write the name of the trained model in the fofn
-        fofn_writer << outname << "\n";
     }
 }
 
@@ -823,26 +862,30 @@ int methyltrain_main(int argc, char** argv)
     parse_methyltrain_options(argc, argv);
     omp_set_num_threads(opt::num_threads);
 
-    Fast5Map name_map(opt::reads_file);
-    
-    // copy the input model into a new type that will hold the trained models
-    const PoreModelMap& input_models = PoreModelSet::get_models(opt::initial_model_type);
-    for(auto model_iter : input_models) {
-        PoreModel model_copy = model_iter.second;
-        model_copy.type = opt::trained_model_type;
-        PoreModelSet::insert_model(model_copy.type, model_copy);
-    }
+    ReadDB read_db;
+    read_db.load(opt::reads_file);
 
-    // Set the alphabet for this run to be the alphabet for the first model
-    assert(!PoreModelSet::get_models(opt::initial_model_type).empty());
-    mtrain_alphabet = PoreModelSet::get_models(opt::initial_model_type).begin()->second.pmalphabet;
+    // Import the models to train into the pore model set
+    assert(!opt::models_fofn.empty());
+    std::vector<const PoreModel*> imported_models = PoreModelSet::initialize(opt::models_fofn);
+    assert(!imported_models.empty());
+
+    // Grab one of the pore models to extract the kit name from (they should all have the same one)
+    const PoreModel& tmp_model = *imported_models.front();
+
+    std::string training_kit = tmp_model.metadata.get_kit_name();
+    mtrain_alphabet = tmp_model.pmalphabet;
+    size_t training_k = tmp_model.k;
+    fprintf(stderr, "Training %s for alphabet %s for %zu-mers\n", training_kit.c_str(), mtrain_alphabet->get_name(), training_k);
 
     for(size_t round = 0; round < opt::num_training_rounds; round++) {
         fprintf(stderr, "Starting round %zu\n", round);
-        train_one_round(name_map, round);
+        train_one_round(read_db, training_kit, mtrain_alphabet->get_name(), training_k, round);
+        /*
         if(opt::write_models) {
-            write_models(PoreModelSet::get_models(opt::trained_model_type), round);
+            write_models(training_kit, mtrain_alphabet->get_name(), training_k, round);
         }
+        */
     }
     return EXIT_SUCCESS;
 }
