@@ -18,6 +18,8 @@
 #include "nanopolish_alphabet.h"
 #include "nanopolish_emissions.h"
 #include "nanopolish_profile_hmm.h"
+#include "nanopolish_pore_model_set.h"
+#include "nanopolish_variant_db.h"
 #include "training_core.hpp"
 #include "invgauss.hpp"
 #include "logger.hpp"
@@ -272,6 +274,109 @@ TEST_CASE( "math", "[math]") {
     REQUIRE( log_normal_pdf(2.25, params) == Approx(log(normal_pdf(2.25, params))) );
 }
 
+TEST_CASE( "scalings", "[scalings]") {
+
+    SquiggleRead test_read;
+    size_t strand = 0;
+    test_read.base_model[0] = PoreModelSet::get_model("r9.4_450bps", "nucleotide", "template", 6);
+    const PoreModel* pore_model = test_read.base_model[0];
+    test_read.scalings[strand].set4(10.0f, 1.2, 0.5, 1.3);
+    
+    assert(pore_model != NULL);
+
+    // Generate events from the pore model
+    const SquiggleScalings& scalings = test_read.scalings[strand];
+    size_t n_events = 100;
+    double duration = 10.0 / 4000.0f;
+    size_t rank = 100;
+    std::default_random_engine generator;
+
+    std::vector<float> lp_truth;
+    std::vector<float> z_truth;
+
+    for(size_t i = 0; i < n_events; ++i) {
+
+        PoreModelStateParams params = pore_model->states[rank];
+        SquiggleEvent event;
+        event.stdv = 1.0f; //unused
+        event.start_time = (i * duration);
+        event.duration = duration;
+        
+        // Generate level from N(a + bu + ct, (do)^2)
+        float g_mean = scalings.shift + scalings.scale * params.level_mean + event.start_time * scalings.drift;
+        float g_stdv = scalings.var * params.level_stdv;
+
+        std::normal_distribution<float> distribution(g_mean, g_stdv);
+        event.mean = distribution(generator);
+        test_read.events[0].push_back(event);
+
+        GaussianParameters gp = { g_mean, g_stdv };
+        lp_truth.push_back(log_normal_pdf(event.mean, gp));
+        z_truth.push_back((event.mean - gp.mean) / gp.stdv);
+    }
+
+    // Calculate log-probability of observing an event by scaling the event
+    for(size_t i = 0; i < n_events; ++i) {
+        float lp = log_probability_match_r9(test_read, *pore_model, rank, i, strand);
+        float z = z_score(test_read, *pore_model, rank, i, strand);
+        REQUIRE( lp == Approx(lp_truth[i]) );
+        REQUIRE( z == Approx(z_truth[i]) );
+    }
+}
+
+size_t factorial(size_t n)
+{
+    if(n == 0 || n == 1) {
+        return 1;
+    }
+
+    size_t out = 1;
+    for(size_t i = 2; i <= n; ++i) {
+        out *= i;
+    }
+    return out;
+}
+
+void test_combinations(size_t n, size_t k, CombinationOption option, std::vector<std::string> expected)
+{
+    Combinations c(n, k, option);
+    size_t idx = 0;
+    while(!c.done()) {
+        std::string cs = c.get_as_string();
+        REQUIRE(cs == expected[idx]);
+        idx++;
+        c.next();
+    }
+    REQUIRE(idx == expected.size());
+}
+
+TEST_CASE( "combinations", "[combinations]") {
+    test_combinations(1, 1, CO_WITHOUT_REPLACEMENT, {"0"});
+    test_combinations(2, 1, CO_WITHOUT_REPLACEMENT, { "0", "1" });
+    test_combinations(2, 2, CO_WITHOUT_REPLACEMENT, { "0,1" });
+    test_combinations(3, 2, CO_WITHOUT_REPLACEMENT, { "0,1", "0,2", "1,2" });
+    test_combinations(4, 4, CO_WITHOUT_REPLACEMENT, { "0,1,2,3", });
+    REQUIRE(factorial(4) == 24);
+
+    // Count we see the right number of combinations
+    size_t n = 10;
+    size_t k = 4;
+    size_t n_comb = factorial(n) / (factorial(k) * factorial(n - k));
+    size_t count = 0;
+    Combinations c(n, k);
+    while(!c.done()) {
+        count++;
+        c.next();
+    }
+    REQUIRE(count == n_comb);
+
+    // With replacement
+    test_combinations(1, 1, CO_WITH_REPLACEMENT, {"0"});
+    test_combinations(2, 1, CO_WITH_REPLACEMENT, { "0", "1" });
+    test_combinations(2, 2, CO_WITH_REPLACEMENT, { "0,0", "0,1", "1,1" });
+    test_combinations(3, 2, CO_WITH_REPLACEMENT, { "0,0", "0,1", "0,2", "1,1", "1,2", "2,2"});
+}
+
 std::string event_alignment_to_string(const std::vector<HMMAlignmentState>& alignment)
 {
     std::string out;
@@ -281,16 +386,82 @@ std::string event_alignment_to_string(const std::vector<HMMAlignmentState>& alig
     return out;
 }
 
+TEST_CASE( "hmm", "[hmm]") {
+
+    // load the FAST5
+    std::string read_id = "01234567-0123-0123-0123-0123456789ab:2D_000:2d";
+    std::string fast5_path = "test/data/LomanLabz_PC_Ecoli_K12_R7.3_2549_1_ch8_file30_strand.fast5";
+
+    ReadDB read_db;
+    read_db.add_signal_path(read_id, fast5_path);
+
+    SquiggleRead sr(read_id, read_db);
+
+    // The reference sequence to align to:
+    std::string ref_subseq = "ATCAGTAAAATAACGTAGAGCGGTAACCTTGCCATAAAGGTCGAGTTTA"
+                             "TTACCATCCTTGTTATAGACTTCGGCAGCGTGTGCTACGTTCGCAGCT";
+
+    // Generate a HMMInputData structure to tell the HMM
+    // which part of the read to align
+    HMMInputData input[2];
+    
+    // template strand
+    input[0].read = &sr;
+    input[0].event_start_idx = 3;
+    input[0].event_stop_idx = 88;
+    input[0].event_stride = 1;
+    input[0].rc = false;
+    input[0].strand = 0;
+    input[0].pore_model = sr.get_model(0, "nucleotide");
+
+    // complement strand
+    input[1].read = &sr;
+    input[1].event_start_idx = 6788;
+    input[1].event_stop_idx = 6697;
+    input[1].event_stride = -1;
+    input[1].rc = true;
+    input[1].strand = 1;
+    input[1].pore_model = sr.get_model(1, "nucleotide");
+    
+    // expected output
+    std::string expected_alignment[2];
+    expected_alignment[0] = 
+        "MMMMMEMKMKMMMMMMMKMMMKMMMKMMMMMMMMMKKMMEEEMMMMMMKMMMM" 
+        "MMMKMMMMMKMKMKMEMKKMKMKKMMMMMMEMMMMKMKMEEMMMMKMEEEEEM";
+
+    expected_alignment[1] = 
+        "MMKMMMKMEEMMKMKMKMEMMMKMMMKMEMMMKMMMKMMMMMMMMMKKMEMMM"
+        "EMMMMMMMMMKMKKMMMMMMMEMMMMMKMMMMMKMEMMMMMKMMMMMEEEEEEEEM";
+
+    double expected_viterbi_last_state[2] = { -237.7808380127, -267.9027709961 };
+    double expected_forward[2] = { -216.053604126, -254.5881347656 };
+
+    for(int si = 0; si <= 1; ++si) {
+
+        // viterbi align
+        std::vector<HMMAlignmentState> event_alignment = profile_hmm_align(ref_subseq, input[si]);
+        std::string ea_str = event_alignment_to_string(event_alignment);
+    
+        // check
+        REQUIRE( ea_str == expected_alignment[si]);
+        REQUIRE( event_alignment.back().l_fm == Approx(expected_viterbi_last_state[si]));
+
+        // forward algorithm
+        double lp = profile_hmm_score(ref_subseq, input[si]);
+        REQUIRE(lp == Approx(expected_forward[si]));
+    }
+}
+
 std::vector< StateTrainingData >
 generate_training_data(const ParamMixture& mixture, size_t n_data,
-                       const std::array< float, 2 >& read_var_rg = { .5f, 1.5f },
+                       const std::array< float, 2 >& scaled_read_var_rg = { .5f, 1.5f },
                        const std::array< float, 2 >& read_scale_sd_rg = { .5f, 1.5f },
                        const std::array< float, 2 >& read_var_sd_rg = { .5f, 1.5f })
 {
     // check parameter sizes
     size_t n_components = mixture.log_weights.size();
     assert(mixture.params.size() == n_components);
-    assert(read_var_rg[0] < read_var_rg[1]);
+    assert(scaled_read_var_rg[0] < scaled_read_var_rg[1]);
     assert(read_scale_sd_rg[0] < read_scale_sd_rg[1]);
     assert(read_var_sd_rg[0] < read_var_sd_rg[1]);
     // set random seed
@@ -320,9 +491,9 @@ generate_training_data(const ParamMixture& mixture, size_t n_data,
         size_t j = discrete_dist(weights.begin(), weights.end())(rg);
         ++population_size[j];
         assert(j < n_components);
-        // draw read_var
-        data[i].read_var = uniform_dist(read_var_rg[0], read_var_rg[1])(rg);
-        data[i].log_read_var = std::log(data[i].read_var);
+        // draw scaled_read_var
+        data[i].scaled_read_var = uniform_dist(scaled_read_var_rg[0], scaled_read_var_rg[1])(rg);
+        data[i].log_scaled_read_var = std::log(data[i].scaled_read_var);
         // draw read_scale_sd
         data[i].read_scale_sd = uniform_dist(read_scale_sd_rg[0], read_scale_sd_rg[1])(rg);
         data[i].log_read_scale_sd = std::log(data[i].read_scale_sd);
@@ -331,8 +502,8 @@ generate_training_data(const ParamMixture& mixture, size_t n_data,
         data[i].log_read_var_sd = std::log(data[i].read_var_sd);
         // scale the state
         auto scaled_params = mixture.params[j];
-        scaled_params.level_stdv *= data[i].read_var;
-        scaled_params.level_log_stdv += data[i].log_read_var;
+        scaled_params.level_stdv *= data[i].scaled_read_var;
+        scaled_params.level_log_stdv += data[i].log_scaled_read_var;
         scaled_params.sd_lambda *= data[i].read_var_sd / data[i].read_scale_sd;
         scaled_params.sd_log_lambda += data[i].log_read_var_sd - data[i].log_read_scale_sd;
         // draw level_mean & level_stdv
@@ -346,7 +517,7 @@ generate_training_data(const ParamMixture& mixture, size_t n_data,
             << "data " << i << " " << j << " "
             << data[i].level_mean << " "
             << data[i].level_stdv << " "
-            << data[i].read_var << " "
+            << data[i].scaled_read_var << " "
             << data[i].read_scale_sd << " "
             << data[i].read_var_sd << std::endl;
     }
@@ -398,76 +569,4 @@ TEST_CASE("training", "[training]")
         CHECK( out_mixture.params[0].level_mean == Approx( um_params.level_mean ).epsilon(.05) );
         CHECK( out_mixture.params[1].level_mean == Approx( um_params.level_mean + delta_level_mean ).epsilon(.05) );
     }
-
-    // next, we test inverse gaussian training for the case where the gaussians are distinct
-    SECTION("inverse_gaussian_1")
-    {
-        float delta_um_rate = .05;
-        float delta_level_mean = 10.0;
-        float delta_sd_mean = 0.3;
-        ParamMixture gen_mixture;
-        gen_mixture.log_weights.push_back(std::log(um_rate + delta_um_rate));
-        gen_mixture.log_weights.push_back(std::log(1 - (um_rate + delta_um_rate)));
-        gen_mixture.params.push_back(um_params);
-        gen_mixture.params.push_back(um_params);
-        gen_mixture.params[1].level_mean += delta_level_mean;
-        gen_mixture.params[1].sd_mean += delta_sd_mean;
-        gen_mixture.params[1].update_sd_stdv();
-        gen_mixture.params[1].update_logs();
-        auto data = generate_training_data(gen_mixture, n_data);
-        ParamMixture in_mixture;
-        in_mixture.log_weights.push_back(std::log(um_rate));
-        in_mixture.log_weights.push_back(std::log(1 - um_rate));
-        in_mixture.params.push_back(um_params);
-        in_mixture.params.push_back(um_params);
-        // encourage the second component to capture points not well fit by the first
-        in_mixture.params[1].level_stdv += 1.0;
-        in_mixture.params[1].update_logs();
-        auto mid_mixture = train_gaussian_mixture(data, in_mixture);
-        CHECK( std::exp(mid_mixture.log_weights[0]) == Approx( um_rate + delta_um_rate ).epsilon(.05) );
-        CHECK( mid_mixture.params[0].level_mean == Approx( um_params.level_mean ).epsilon(.05) );
-        CHECK( mid_mixture.params[1].level_mean == Approx( um_params.level_mean + delta_level_mean ).epsilon(.05) );
-        auto out_mixture = train_invgaussian_mixture(data, mid_mixture);
-        CHECK( std::exp(out_mixture.log_weights[0]) == Approx( um_rate + delta_um_rate ).epsilon(.05) );
-        CHECK( out_mixture.params[0].level_mean == Approx( um_params.level_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[1].level_mean == Approx( um_params.level_mean + delta_level_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[0].sd_mean == Approx( um_params.sd_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[1].sd_mean == Approx( um_params.sd_mean + delta_sd_mean ).epsilon(.05) );
-    }
-
-    // next, we test inverse gaussian training for the case where the gaussians are very similar
-    SECTION("inverse_gaussian_2")
-    {
-        float delta_um_rate = .05;
-        float delta_level_mean = 2.0;
-        float delta_sd_mean = 0.3;
-        ParamMixture gen_mixture;
-        gen_mixture.log_weights.push_back(std::log(um_rate + delta_um_rate));
-        gen_mixture.log_weights.push_back(std::log(1 - (um_rate + delta_um_rate)));
-        gen_mixture.params.push_back(um_params);
-        gen_mixture.params.push_back(um_params);
-        gen_mixture.params[1].level_mean += delta_level_mean;
-        gen_mixture.params[1].sd_mean += delta_sd_mean;
-        gen_mixture.params[1].update_sd_stdv();
-        gen_mixture.params[1].update_logs();
-        auto data = generate_training_data(gen_mixture, n_data);
-        ParamMixture in_mixture;
-        in_mixture.log_weights.push_back(std::log(um_rate));
-        in_mixture.log_weights.push_back(std::log(1 - um_rate));
-        in_mixture.params.push_back(um_params);
-        in_mixture.params.push_back(um_params);
-        // encourage the second component to capture points not well fit by the first
-        in_mixture.params[1].level_stdv += 1.0;
-        in_mixture.params[1].update_logs();
-        auto mid_mixture = train_gaussian_mixture(data, in_mixture);
-        CHECK( std::exp(mid_mixture.log_weights[0]) == Approx( um_rate + delta_um_rate ).epsilon(.05) );
-        CHECK( mid_mixture.params[0].level_mean == Approx( um_params.level_mean ).epsilon(.05) );
-        CHECK( mid_mixture.params[1].level_mean == Approx( um_params.level_mean + delta_level_mean ).epsilon(.05) );
-        auto out_mixture = train_invgaussian_mixture(data, mid_mixture);
-        CHECK( std::exp(out_mixture.log_weights[0]) == Approx( um_rate + delta_um_rate ).epsilon(.05) );
-        CHECK( out_mixture.params[0].level_mean == Approx( um_params.level_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[1].level_mean == Approx( um_params.level_mean + delta_level_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[0].sd_mean == Approx( um_params.sd_mean ).epsilon(.05) );
-        CHECK( out_mixture.params[1].sd_mean == Approx( um_params.sd_mean + delta_sd_mean ).epsilon(.05) );
-    }
 }
